#include <stdio.h>
#include <stdlib.h>

#include <jmp.h>
#include <cls.h>
#include <method.h>
#include <monitor_ops.h>

/* This should be a coverage profiler once it is finished. 
 * It should get all classes and all methods and when the 
 * system shuts down it should report all classes.methods
 * that have (not) been visited. 
 * This is something that is very useful when one tries to 
 * write tests for a large system. If the tests give full 
 * coverage you can be fairly certain that the test suite
 * is quite good. 
 */

/** This is the interface into the jvm */
JVMPI_Interface* jvmpi;
static hashtab* classes;
static hashtab* methods;
static int down = 0;   /** since we get multiple shutdowns... */

static void lock_all () {
    jmphash_lock (classes);
    jmphash_lock (methods);
}

static void unlock_all () {
    jmphash_unlock (methods);
    jmphash_unlock (classes);
}

/** Get a class pointer from its id. */
cls* get_class (const jobjectID class_id) {
    cls c;
    cls* cp = NULL;
    cls_set_class_id (&c, class_id);
    cp = jmphash_search (&c, classes);
    return cp;
}

/** Get the hashtable of all loaded classes. */
hashtab* get_classes () {
    return classes;
}

/** Get a method object pointer from its id. */
method* get_method (const jmethodID method_id) {
    method m;
    method* mp = NULL;
    method_set_method_id (&m, method_id);
    mp = jmphash_search (&m, methods);
    return mp;
}

/** A class has been loaded. 
 *  Locks will only be taken if requested = 0.
 *  Requested class_load are assumed to be locked already..
 */
static void class_load (JNIEnv* env,
			const char *class_name, 
			const char *source_name, 
			jint num_interfaces,
			jint num_methods, 
			JVMPI_Method *meths, 
			jint num_static_fields,
			JVMPI_Field *statics, 
			jint num_instance_fields, 
			JVMPI_Field *instances, 
			jobjectID class_id,
			jint requested) {
    cls* c;
    method* m;
    int i;
    if (!classes)
	return;
    if (!requested)
	jmphash_lock (classes);
    c = get_class (class_id);
    if (c == NULL) {
	c = cls_new (class_name, source_name, class_id, num_interfaces, 
		     num_static_fields, statics, 
		     num_instance_fields, instances);
	if (c == NULL) {
	    fprintf (stderr, "class_load: failed to allocate cls: %s, %s, %p.\n", 
		     class_name, source_name, class_id);
	} else {
	    jmphash_insert (c, classes);
	}
	
	/* and its methods */
	if (!requested)
	    jmphash_lock (methods);
	
	for (i = 0; i < num_methods; i++) {
	    JVMPI_Method *jmethod;
	    jmethod = meths + i;
	    m = get_method (jmethod->method_id);
	    if (m == NULL) {
		m = method_new (jmethod->method_name, jmethod->method_signature,
				jmethod->start_lineno, jmethod->end_lineno, 
				jmethod->method_id, c);
		if (m == NULL) {
		    fprintf (stderr, "class_load: failed to allocate method: %s, %p, %s, %s.\n",
			     class_name, class_id, jmethod->method_name, jmethod->method_signature);
		} else {
		    jmphash_insert (m, methods);
		}
	    }
	}

	if (!requested)
	    jmphash_unlock (methods);
    }
    
    if (!requested)
	jmphash_unlock (classes);
}

/** A method is beeing called. */
static void method_entry (jmethodID method_id, JNIEnv *env) {
    method* m;
    jmphash_lock (methods);
    m = get_method (method_id);
    if (m != NULL)
	m->calls = 1;
    jmphash_unlock(methods);
}

/** Do initial setup. */
static int setup () {
    classes = jmphash_new (2000, cls_jmphash_func, cls_cmp_func);
    methods = jmphash_new (10000, method_jmphash_func, method_cmp_func);
    return 0;
}

static void disable_events (int keep_threads) {
    /* Disable standard events */
    jvmpi->DisableEvent (JVMPI_EVENT_JVM_SHUT_DOWN, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_CLASS_LOAD, NULL);
    jvmpi->DisableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
}

static void print_uncalled_methods_class (void* mm, void* cc) {
    method* m = (method*)mm;
    cls* c = (cls*)cc;
    if (method_get_owner (m) != c)
	return;
    if (m->calls == 0) 
	fprintf (stdout, "%s: %s\n", cls_get_name (c), method_get_method_jmpname (m));
}

static void print_uncalled_methods (void* cc) {
    jmphash_for_each_with_arg ((jmphash_iter_fa)print_uncalled_methods_class, methods, cc);
}

/** Tear down all allocations we have made. 
 */
static int teardown () {
    jmphash_for_each ((jmphash_iter_f)print_uncalled_methods, classes);

    if (methods) {
	jmphash_for_each ((jmphash_iter_f)method_free, methods);
	jmphash_free (methods);
	methods = NULL;
    }
    if (classes) {
	jmphash_for_each ((jmphash_iter_f)cls_free, classes);
	jmphash_free (classes);
	classes = NULL;
    }
    down = 1;
    return 0;
}

/** The jvm is shutting down so clean up our mess.
 */
static void jvm_shut_down () {
    if (down)
	return;

    disable_events (1);    
    fprintf (stdout, "jvm_shut_down.\n");
    lock_all ();
    teardown ();
    unlock_all ();
}

static void enable_events () {
    /* Enable standard events */
    jvmpi->EnableEvent (JVMPI_EVENT_JVM_SHUT_DOWN, NULL);    
    jvmpi->EnableEvent (JVMPI_EVENT_CLASS_LOAD, NULL);
    jvmpi->EnableEvent (JVMPI_EVENT_METHOD_ENTRY, NULL);
}

/** Switch on the event_type and call the right method. 
 */
void notifyEvent (JVMPI_Event *e) {
    switch(e->event_type) {
    case JVMPI_EVENT_JVM_SHUT_DOWN:
	jvm_shut_down ();
	break;
    case JVMPI_EVENT_CLASS_LOAD:
    case JVMPI_EVENT_CLASS_LOAD | JVMPI_REQUESTED_EVENT:
	class_load (e->env_id,
		    e->u.class_load.class_name, e->u.class_load.source_name,
		    e->u.class_load.num_interfaces, e->u.class_load.num_methods,
		    e->u.class_load.methods, e->u.class_load.num_static_fields,
		    e->u.class_load.statics, e->u.class_load.num_instance_fields, 
		    e->u.class_load.instances, e->u.class_load.class_id, 
		    e->event_type & JVMPI_REQUESTED_EVENT);
	break;
    case JVMPI_EVENT_METHOD_ENTRY:
	method_entry (e->u.method.method_id, e->env_id);
	break;
    default:
	break;
    }
}

/** This method is the one that is called by the jvm on startup. 
 */
JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *jvm, char *options, void *reserved) {
    fprintf (stdout, "jmpcoverage/%s initializing: (%s):...\n", VERSION, (options ? options : ""));
    /* get jvmpi interface pointer */
    if (((*jvm)->GetEnv(jvm, (void **)&jvmpi, JVMPI_VERSION_1)) < 0) {
	fprintf (stderr, "jmp: error in obtaining jvmpi interface pointer\n");
	return JNI_ERR;
    }

    jvmpi->NotifyEvent = notifyEvent;   
    if (setup ())
	return JNI_ERR;
    fprintf (stdout, "jmpcoverage: Enabling localization.\n");
    /* Enable internationalization */
#ifdef LC_ALL    
    setlocale (LC_ALL, "");
    bindtextdomain (PACKAGE, LOCALEDIR);
    bind_textdomain_codeset (PACKAGE,"UTF-8");
    textdomain (PACKAGE);
#endif
    fprintf (stdout, _("jmpcoverage: Loaded and registered correctly.\n"));

    /* initialize jvmpi interface */
    enable_events ();
    return JNI_OK;
}

